; AUTOMATICALLY GENERATED -- DO NOT EDIT
(define Incoming (meep-in-or-out-Incoming))
(define Outgoing (meep-in-or-out-Outgoing))
(define CONNECT-PHASE (meep-connect-phase-CONNECT-PHASE))
(define CONNECT-NEGATE (meep-connect-phase-CONNECT-NEGATE))
(define CONNECT-COPY (meep-connect-phase-CONNECT-COPY))
(define Periodic (meep-boundary-condition-Periodic))
(define Metallic (meep-boundary-condition-Metallic))
(define Magnetic (meep-boundary-condition-Magnetic))
(define None (meep-boundary-condition-None))
(define Connecting (meep-time-sink-Connecting))
(define Stepping (meep-time-sink-Stepping))
(define Boundaries (meep-time-sink-Boundaries))
(define MpiAllTime (meep-time-sink-MpiAllTime))
(define MpiOneTime (meep-time-sink-MpiOneTime))
(define FieldOutput (meep-time-sink-FieldOutput))
(define FourierTransforming (meep-time-sink-FourierTransforming))
(define MPBTime (meep-time-sink-MPBTime))
(define GetFarfieldsTime (meep-time-sink-GetFarfieldsTime))
(define Other (meep-time-sink-Other))
(define FieldUpdateB (meep-time-sink-FieldUpdateB))
(define FieldUpdateH (meep-time-sink-FieldUpdateH))
(define FieldUpdateD (meep-time-sink-FieldUpdateD))
(define FieldUpdateE (meep-time-sink-FieldUpdateE))
(define BoundarySteppingB (meep-time-sink-BoundarySteppingB))
(define BoundarySteppingWH (meep-time-sink-BoundarySteppingWH))
(define BoundarySteppingPH (meep-time-sink-BoundarySteppingPH))
(define BoundarySteppingH (meep-time-sink-BoundarySteppingH))
(define BoundarySteppingD (meep-time-sink-BoundarySteppingD))
(define BoundarySteppingWE (meep-time-sink-BoundarySteppingWE))
(define BoundarySteppingPE (meep-time-sink-BoundarySteppingPE))
(define BoundarySteppingE (meep-time-sink-BoundarySteppingE))
(define Ex (meep-component-Ex))
(define Ey (meep-component-Ey))
(define Er (meep-component-Er))
(define Ep (meep-component-Ep))
(define Ez (meep-component-Ez))
(define Hx (meep-component-Hx))
(define Hy (meep-component-Hy))
(define Hr (meep-component-Hr))
(define Hp (meep-component-Hp))
(define Hz (meep-component-Hz))
(define Dx (meep-component-Dx))
(define Dy (meep-component-Dy))
(define Dr (meep-component-Dr))
(define Dp (meep-component-Dp))
(define Dz (meep-component-Dz))
(define Bx (meep-component-Bx))
(define By (meep-component-By))
(define Br (meep-component-Br))
(define Bp (meep-component-Bp))
(define Bz (meep-component-Bz))
(define Dielectric (meep-component-Dielectric))
(define Permeability (meep-component-Permeability))
(define NO-COMPONENT (meep-component-NO-COMPONENT))
(define Sx (meep-derived-component-Sx))
(define Sy (meep-derived-component-Sy))
(define Sr (meep-derived-component-Sr))
(define Sp (meep-derived-component-Sp))
(define Sz (meep-derived-component-Sz))
(define EnergyDensity (meep-derived-component-EnergyDensity))
(define D-EnergyDensity (meep-derived-component-D-EnergyDensity))
(define H-EnergyDensity (meep-derived-component-H-EnergyDensity))
(define D1 (meep-ndim-D1))
(define D2 (meep-ndim-D2))
(define D3 (meep-ndim-D3))
(define Dcyl (meep-ndim-Dcyl))
(define E-stuff (meep-field-type-E-stuff))
(define H-stuff (meep-field-type-H-stuff))
(define D-stuff (meep-field-type-D-stuff))
(define B-stuff (meep-field-type-B-stuff))
(define PE-stuff (meep-field-type-PE-stuff))
(define PH-stuff (meep-field-type-PH-stuff))
(define WE-stuff (meep-field-type-WE-stuff))
(define WH-stuff (meep-field-type-WH-stuff))
(define High (meep-boundary-side-High))
(define Low (meep-boundary-side-Low))
(define X (meep-direction-X))
(define Y (meep-direction-Y))
(define Z (meep-direction-Z))
(define R (meep-direction-R))
(define P (meep-direction-P))
(define NO-DIRECTION (meep-direction-NO-DIRECTION))
(define Centered Dielectric)
