#!/usr/bin/env python3
from distutils.core import setup

setup(
    name="python-markdown-math",
    description="Math extension for Python-Markdown",
    author="Dmitry Shachnev",
    author_email="mitya57@gmail.com",
    version="0.1",
    url="https://github.com/mitya57/python-markdown-math",
    py_modules=["mdx_math"],
    license="BSD",
)
